import re

from mal import config
from mal.base import _Base


class _SearchResult:
    def __init__(self, mal_id, url, image_url, title, synopsis, media_type, score):
        self.mal_id = mal_id
        self.url = url
        self.image_url = image_url
        self.title = title
        self.synopsis = synopsis
        self.type = media_type
        self.score = score


class _Search(_Base):
    def __init__(self, query, mal_type, timeout):
        if len(query) > 100:
            raise ValueError("Query cannot be more than 100 characters")
        super().__init__(timeout)
        self._query = query
        self._url = config.MAL_ENDPOINT + "{}.php?q={}".format(mal_type, query)
        self._page = self._parse_url(self._url)
        self._inner_page = self._page.find("div", {"class": "js-block-list"})
        if self._inner_page is None:
            raise ValueError("No results found")

    @staticmethod
    def _parse_mal_id(url):
        pattern = r"^https:\/\/myanimelist\.net\/.+\/(\d+)\/.+$"
        match = re.match(pattern, url)
        try:
            return int(match.group(1))
        except ValueError:
            return None

    @staticmethod
    def _parse_eps_vols(text):
        text = text.strip()
        try:
            return int(text)
        except ValueError:
            return None

    @staticmethod
    def _parse_score(text):
        text = text.strip()
        try:
            return float(text)
        except ValueError:
            return None

    @staticmethod
    def _remove_suffix(text, suffix):
        if text.endswith(suffix):
            return text[:-len(suffix)]
        return text



MAL_ENDPOINT = "https://myanimelist.net/"
NOT_FOUND_TITLE = "404 Not Found - MyAnimeList.net "
NO_BACKGROUND_INFO = "No background information has been added to this title. Help improve our database by adding background information here."

TIMEOUT = 5



from mal import config
from mal.search import _Search, _SearchResult


class AnimeSearchResult(_SearchResult):
    def __init__(self, mal_id, url, image_url, title, synopsis, media_type, episodes, score):
        super().__init__(mal_id, url, image_url, title, synopsis, media_type, score)
        self.episodes = episodes


class AnimeSearch(_Search):
    def __init__(self, query, timeout=config.TIMEOUT):
        super().__init__(query, "anime", timeout)

    def reload(self):
        self.__init__(self._query)

    @property
    def results(self):
        try:
            self._results
        except AttributeError:
            trs = self._inner_page.find_all("tr")
            results = []
            for tr in trs[1:]:
                tds = tr.find_all("td")

                url = tds[0].find("a")["href"]
                mal_id = self._parse_mal_id(url)

                results.append(AnimeSearchResult(
                    mal_id=mal_id,
                    url=url,
                    image_url=tds[0].find("img")["data-src"],
                    title=tds[1].find("strong").text.strip(),
                    synopsis=self._remove_suffix(tds[1].find("div", {"class": "pt4"}).text.strip(), "read more."),
                    media_type=tds[2].text.strip(),
                    episodes=self._parse_eps_vols(tds[3].text),
                    score=self._parse_score(tds[4].text)
                ))
            self._results = results
        return self._results



import requests
from bs4 import BeautifulSoup


class _Base:
    def __init__(self, timeout):
        self.timeout = timeout

    def _parse_url(self, url):
        page_response = requests.get(url, timeout=self.timeout)
        return BeautifulSoup(page_response.content, "html.parser")



from mal.mal import _MAL

from mal import config


class Manga(_MAL):
    def __init__(self, mal_id, timeout=config.TIMEOUT):
        super().__init__(mal_id, "manga", timeout)

    def reload(self):
        self.__init__(self._mal_id)

    @property
    def volumes(self):
        try:
            self._volumes
        except AttributeError:
            self._volumes = self._get_span_text(self._border_spans, "Volumes:", int)
        return self._volumes

    @property
    def chapters(self):
        try:
            self._chapters
        except AttributeError:
            self._chapters = self._get_span_text(self._border_spans, "Chapters:", int)
        return self._chapters

    @property
    def published(self):
        try:
            self._published
        except AttributeError:
            self._published = self._get_span_text(self._border_spans, "Published:", str)
        return self._published

    @property
    def authors(self):
        try:
            self._authors
        except AttributeError:
            self._authors = self._get_span_text(self._border_spans, "Authors:", list)
        return self._authors

    @property
    def related_manga(self):
        try:
            self._related_manga
        except AttributeError:
            self._related_manga = self._get_related()
        return self._related_manga

    @property
    def background(self):
        try:
            self._background
        except AttributeError:
            self._background = self._parse_background("h2")
        return self._background



import re

from mal import config
from mal.base import _Base


class _MAL(_Base):
    def __init__(self, mal_id, mal_type, timeout):
        super().__init__(timeout)
        self._mal_id = mal_id
        self._url = config.MAL_ENDPOINT + "{}/{}".format(mal_type, mal_id)
        self._page = self._parse_url(self._url)
        title = self._page.find("meta", property="og:title")["content"]
        if title == config.NOT_FOUND_TITLE:
            raise ValueError("No such id on MyAnimeList")
        else:
            self._title = title
            url = self._page.find("meta", property="og:url")["content"]
            self._url = url
            self._page_stats = self._parse_url(url + "/stats")
            self._border_spans = self._page.find("td", {"class": "borderClass"}).findChildren("span",
                                                                                              {"class": "dark_text"})

    @staticmethod
    def _get_span_text(page, key, typing, bypass_link=False):
        for span in page:
            if span.get_text() == key:
                first_link = span.parent.a
                if first_link and not bypass_link:
                    if typing == str:
                        return first_link.text
                    if typing == list:
                        res = [link.get_text() for link in span.parent.findChildren("a")]
                        if "add some" in res:
                            res.remove("add some")
                        return res
                else:
                    results = span.parent.findChildren(text=True, recursive=True)
                    result = results[results.index(key) + 1].strip()
                    if result == "Unknown":
                        return None
                    else:
                        if typing == int:
                            num = re.sub("[^0-9]", "", result)
                            if num:
                                return int(num)
                            else:
                                return None
                        if typing == str:
                            return result
                        if typing == list:
                            return [element.strip() for element in result.split(",")]
        if typing == list:
            return []
        return None

    @staticmethod
    def _get_itemprop_value(page, itemprop, element, typing):
        result = page.find(element, itemprop=itemprop)
        if result is not None:
            if typing == int or typing == float:
                return typing(re.sub("[^0-9.]", "", result.text))
            elif typing == str:
                return result.text.replace("\n", " ").replace("\r", "").strip()
        else:
            return None

    def _get_related(self):
        data = {}
        rows = self._page.find("td", {"class": "pb24"}).table.findChildren("tr")
        for row in rows:
            data[row.td.text[:-1]] = [link.get_text() for link in row.findChildren("a")]
        return data

    def _parse_background(self, element):
        raw_string = self._page.find(element, {"style": "margin-top: 15px;"}).parent.text
        result = raw_string[raw_string.index("EditBackground") + 14:].replace("\n", " ").replace("\r", "").strip()
        if result == config.NO_BACKGROUND_INFO:
            return None
        return result

    @property
    def mal_id(self):
        return self._mal_id

    @property
    def title(self):
        return self._title

    @property
    def title_english(self):
        try:
            self._title_english
        except AttributeError:
            self._title_english = self._get_span_text(self._border_spans, "English:", str)
        return self._title_english

    @property
    def title_japanese(self):
        try:
            self._title_japanese
        except AttributeError:
            self._title_japanese = self._get_span_text(self._border_spans, "Japanese:", str)
        return self._title_japanese

    @property
    def title_synonyms(self):
        try:
            self._title_synonyms
        except AttributeError:
            self._title_synonyms = self._get_span_text(self._border_spans, "Synonyms:", list)
        return self._title_synonyms

    @property
    def url(self):
        return self._url

    @property
    def image_url(self):
        try:
            self._image_url
        except AttributeError:
            self._image_url = self._page.find("meta", property="og:image")["content"]
        return self._image_url

    @property
    def type(self):
        try:
            self._type
        except AttributeError:
            self._type = self._get_span_text(self._border_spans, "Type:", str)
        return self._type

    @property
    def status(self):
        try:
            self._status
        except AttributeError:
            self._status = self._get_span_text(self._border_spans, "Status:", str)
        return self._status

    @property
    def genres(self):
        try:
            self._genres
        except AttributeError:
            self._genres = self._get_span_text(self._border_spans, "Genres:", list)
        return self._genres

    @property
    def score(self):
        try:
            self._score
        except AttributeError:
            self._score = self._get_itemprop_value(self._page, "ratingValue", "span", float)
        return self._score

    @property
    def scored_by(self):
        try:
            self._scored_by
        except AttributeError:
            self._scored_by = self._get_itemprop_value(self._page, "ratingCount", "span", int)
        return self._scored_by

    @property
    def rank(self):
        try:
            self._rank
        except AttributeError:
            self._rank = self._get_span_text(self._border_spans, "Ranked:", int, True)
        return self._rank

    @property
    def popularity(self):
        try:
            self._popularity
        except AttributeError:
            self._popularity = self._get_span_text(self._border_spans, "Popularity:", int)
        return self._popularity

    @property
    def members(self):
        try:
            self._members
        except AttributeError:
            self._members = self._get_span_text(self._border_spans, "Members:", int)
        return self._members

    @property
    def favorites(self):
        try:
            self._favorites
        except AttributeError:
            self._favorites = self._get_span_text(self._border_spans, "Favorites:", int)
        return self._favorites

    @property
    def synopsis(self):
        try:
            self._synopsis
        except AttributeError:
            self._synopsis = self._get_itemprop_value(self._page, "description", "p", str)
        return self._synopsis



from mal.mal import _MAL

from mal import config


class Anime(_MAL):
    def __init__(self, mal_id, timeout=config.TIMEOUT):
        super().__init__(mal_id, "anime", timeout)

    def reload(self):
        self.__init__(self._mal_id)

    def _get_op_ed(self, option):
        themes = []
        if option == "op":
            data = self._page.find("div", {"class": "opnening"}).parent
        else:
            data = self._page.find("div", {"class": "ending"}).parent
        data = data.findChildren("span", {"class": "theme-song"})
        if data:
            if len(data) > 1:
                for theme in data:
                    themes.append(theme.text[4:])
            else:
                themes = [data[0].text]
            return themes
        else:
            return None

    @property
    def episodes(self):
        try:
            self._episodes
        except AttributeError:
            self._episodes = self._get_span_text(self._border_spans, "Episodes:", int)
        return self._episodes

    @property
    def aired(self):
        try:
            self._aired
        except AttributeError:
            self._aired = self._get_span_text(self._border_spans, "Aired:", str)
        return self._aired

    @property
    def premiered(self):
        try:
            self._premiered
        except AttributeError:
            self._premiered = self._get_span_text(self._border_spans, "Premiered:", str)
        return self._premiered

    @property
    def broadcast(self):
        try:
            self._broadcast
        except AttributeError:
            self._broadcast = self._get_span_text(self._border_spans, "Broadcast:", str)
        return self._broadcast

    @property
    def producers(self):
        try:
            self._producers
        except AttributeError:
            self._producers = self._get_span_text(self._border_spans, "Producers:", list)
        return self._producers

    @property
    def licensors(self):
        try:
            self._licensors
        except AttributeError:
            self._licensors = self._get_span_text(self._border_spans, "Licensors:", list)
        return self._licensors

    @property
    def studios(self):
        try:
            self._studios
        except AttributeError:
            self._studios = self._get_span_text(self._border_spans, "Studios:", list)
        return self._studios

    @property
    def source(self):
        try:
            self._source
        except AttributeError:
            self._source = self._get_span_text(self._border_spans, "Source:", str)
        return self._source

    @property
    def duration(self):
        try:
            self._duration
        except AttributeError:
            self._duration = self._get_span_text(self._border_spans, "Duration:", str)
        return self._duration

    @property
    def rating(self):
        try:
            self._rating
        except AttributeError:
            self._rating = self._get_span_text(self._border_spans, "Rating:", str)
        return self._rating

    @property
    def related_anime(self):
        try:
            self._related_anime
        except AttributeError:
            self._related_anime = self._get_related()
        return self._related_anime

    @property
    def opening_themes(self):
        try:
            self._opening_themes
        except AttributeError:
            self._opening_themes = self._get_op_ed("op")
        return self._opening_themes

    @property
    def ending_themes(self):
        try:
            self._ending_themes
        except AttributeError:
            self._ending_themes = self._get_op_ed("ed")
        return self._ending_themes

    @property
    def background(self):
        try:
            self._background
        except AttributeError:
            self._background = self._parse_background("div")
        return self._background



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mal-api",
    version="0.3.1",
    description="A local MyAnimeList API ",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Daren Liang",
    author_email="darenliang@outlook.com",
    url="https://github.com/darenliang/mal-api",
    packages=setuptools.find_packages(),
    install_requires=["requests", "beautifulsoup4"],
    keywords=["api", "myanimelist"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
